# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym
import numpy as np
from gym import spaces
from numbers import Number
from gym_ignition.utils import logger
from gym_ignition.utils.typing import State, Action, Observation, SeedList


class GymppEnv(gym.Env):
    """Class that exposes C++ Gazebo environments

    This class encapsulates environments created as C++ plugins. Plugins that implement
    the provided gympp and gazebo interfaces are inherently compatible with this
    class if they support the C++ factory.

    The users of this class need to implement the following method:

    _plugin_metadata

    The methods inherited from gym.Env are already implemented.
    """

    metadata = {'render.modes': ['human']}

    def __init__(self):
        # Private attributes
        self._env = None
        self._action_space = None
        self._observation_space = None
        self._act_dt = None
        self._obs_dt = None
        self._np_random = None

        # Seed the environment
        self.seed()

        # Trigger the creation of the spaces
        action_space = self.action_space
        observation_space = self.observation_space
        assert(action_space and observation_space), "Failed to create spaces"

    @property
    def gympp_env(self):
        if self._env:
            return self._env

        import gympp_bindings as bindings

        # Get the metadata
        md = self._plugin_metadata

        # Register the environment
        factory = bindings.GymFactory.Instance()
        factory.register_plugin(md)

        # Load the environment from gympp
        self._env = factory.make(md.get_environment_name())
        assert self._env, "Failed to create environment " + md.get_environment_name()

        # Set the verbosity
        logger.set_level(gym.logger.MIN_LEVEL)

        # Return the gympp environment
        return self._env

    @property
    def gazebo(self):
        import gympp_bindings as bindings
        return bindings.env_to_gazebo_simulator(self.gympp_env)

    @property
    def action_space(self) -> gym.Space:
        if self._action_space:
            return self._action_space

        # Create the action space from the metadata
        md = self._plugin_metadata
        self._action_space, self._act_dt = \
            self._create_space(md.get_action_space_metadata())

        return self._action_space

    @property
    def observation_space(self) -> gym.Space:
        if self._observation_space:
            return self._observation_space

        # Create the action space from the metadata
        md = self._plugin_metadata
        self._observation_space, self._obs_dt = self._create_space(
            md.get_observation_space_metadata())

        return self._observation_space

    def step(self, action: Action) -> State:
        assert self.action_space.contains(action), \
            "The action does not belong to the action space"

        # The bindings do not accept yet numpy types as arguments. We need to covert
        # numpy variables to the closer python type.

        # Check if the input variable is a numpy type
        is_numpy = type(action).__module__ == np.__name__

        if is_numpy:
            if isinstance(action, np.ndarray):
                action = action.tolist()
            elif isinstance(action, np.number):
                action = action.item()
            else:
                assert False

        # Actions must be std::vector objects, so if the passed action is a scalar
        # we have to store it inside a list object before passing it to the bindings
        if isinstance(action, Number):
            action_list = [action]
        else:
            action_list = list(action)

        # Create the gympp::Sample object
        import gympp_bindings as bindings
        action_buffer = getattr(bindings, 'Vector' + self._act_dt)(action_list)
        action_sample = bindings.Sample(action_buffer)

        # Execute the step and get the std::optional<gympp::State> object
        state_optional = self.gympp_env.step(action_sample)
        assert state_optional.has_value()

        # Get the gympp::State
        state = state_optional.value()

        # Get the std::vector buffer of gympp::Observation
        method = 'get_buffer_' + self._obs_dt.lower()
        observation_vector = getattr(state.observation, method)()
        assert observation_vector, "Failed to get the observation buffer"
        assert observation_vector.size() > 0, "The observation does not contain elements"

        # Convert the SWIG type to a list
        observation_list = list(observation_vector)

        # Convert the observation to a numpy array (this is the only required copy)
        if isinstance(self.observation_space, gym.spaces.Box):
            observation = np.array(observation_list)
        elif isinstance(self.observation_space, gym.spaces.Discrete):
            assert observation_vector.size() == 1, "The buffer has the wrong dimension"
            observation = observation_list[0]
        else:
            assert False, "Space not supported"

        assert self.observation_space.contains(observation), \
            "The returned observation does not belong to the space"

        # Create the info dict
        info = {'gympp': state.info}

        # Return the tuple
        return State((observation, state.reward, state.done, info))

    def reset(self) -> Observation:

        # Get std::optional<gympp::Observation>
        obs_optional = self.gympp_env.reset()
        assert obs_optional.has_value(), "The environment didn't return the observation"

        # Get gympp::Observation
        gympp_observation = obs_optional.value()

        # Get the std::vector buffer of gympp::Observation
        method = 'get_buffer_' + self._obs_dt.lower()
        observation_vector = getattr(gympp_observation, method)()
        assert observation_vector, "Failed to get the observation buffer"
        assert observation_vector.size() > 0, "The observation does not contain elements"

        # Convert the SWIG type to a list
        observation_list = list(observation_vector)

        # Convert the observation to a numpy array (this is the only required copy)
        if isinstance(self.observation_space, gym.spaces.Box):
            observation = np.array(observation_list)
        elif isinstance(self.observation_space, gym.spaces.Discrete):
            assert observation_vector.size() == 1, "The buffer has the wrong dimension"
            observation = observation_list[0]
        else:
            assert False, "Space not supported"

        assert self.observation_space.contains(observation), \
            "The returned observation does not belong to the space"

        # Return the list
        return observation

    def render(self, mode: str = 'human') -> None:

        import gympp_bindings as bindings

        render_mode = {'human': bindings.Environment.RenderMode_human}
        ok = self.gympp_env.render(render_mode[mode])

        assert ok, "Failed to render environment"

    def close(self) -> None:
        return

    def seed(self, seed: int = None) -> SeedList:

        if not seed:
            seed = np.random.randint(2**32 - 1)

        # Seed numpy
        self._np_random = np.random
        self._np_random.seed(seed)

        # Seed the environment
        vector_seeds = self.gympp_env.seed(seed)

        # Seed the spaces
        self.action_space.seed(seed)
        self.observation_space.seed(seed)

        return SeedList(list(vector_seeds))

    @property
    def _plugin_metadata(self):
        """Return metadata of the gympp plugin

        Loading an environment created with gympp in python requires the knowledge of
        the plugin metadata that implements it. Every environment should implement this
        method.

        Returns:
            gympp.PluginMetadata: The metadata of the C++ environment plugin
        """
        raise NotImplementedError

    @classmethod
    def _create_space(cls, md=None):
        """Create an object of the gym.space package from gympp space metadata

        Note: In order to map the dynamically typed nature of python to C++, this class
              must know the data type of the gympp buffers to call the right C++ template
              instance stored in the swig bindings. For this reason, this method returns
              also a method suffix string (such as "_d" for double) that is appended to
              the calls of the swig bindings methods (e.g. obs.getBuffer_d()).
        """
        import gympp_bindings as bindings
        assert isinstance(md, bindings.SpaceMetadata), "Wrong type for method argument"

        space_type = md.get_type()
        low = md.get_low_limit()
        high = md.get_high_limit()
        dims = md.get_dimensions()

        if space_type is bindings.SpaceType_box:
            if not dims:
                assert len(low) == len(high), "Sizes of low and high limits do not match"
                return spaces.Box(np.array(low), np.array(high)), "D"
            else:
                assert len(low) == 1, "The size of the limit is not valid"
                assert len(high) == 1, "The size of the limit is not valid"
                return spaces.Box(low[0], high[0], dims), "D"

        elif space_type is bindings.SpaceType_discrete:
            assert len(dims) == 1, "The specified space dimension is not valid"
            return spaces.Discrete(dims[0]), "I"

        else:
            assert False, "This space type is not supported"



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from . import cartpole
from . import gympp_env
from gym.envs import register


register(
    id='CartPoleDiscrete-Gympp-v0',
    max_episode_steps=5000,
    entry_point='gym_ignition.experimental.gympp.cartpole:CartPoleDiscrete')


def get_empty_world() -> str:

    empty_world = f"""<?xml version="1.0" ?>
<sdf version="1.6">
    <world name="default">
        <physics default="true" type="dart">
        </physics>
        <plugin filename="libPhysicsSystem.so"
                name="scenario::plugins::gazebo::Physics">
        </plugin>
        <plugin filename="libECMProvider.so"
                name="scenario::plugins::gazebo::ECMProvider">
        </plugin>
        <plugin filename="libignition-gazebo-scene-broadcaster-system.so"
                name="ignition::gazebo::systems::SceneBroadcaster">
        </plugin>
        <light type="directional" name="sun">
            <cast_shadows>true</cast_shadows>
            <pose>0 0 10 0 0 0</pose>
            <diffuse>1 1 1 1</diffuse>
            <specular>0.5 0.5 0.5 1</specular>
            <attenuation>
                <range>1000</range>
                <constant>0.9</constant>
                <linear>0.01</linear>
                <quadratic>0.001</quadratic>
            </attenuation>
            <direction>-0.5 0.1 -0.9</direction>
        </light>
    </world>
</sdf>"""

    return empty_world



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from . import math
from . import misc
from .typing import *
from . import logger
from . import scenario
from . import resource_finder
from . import inverse_kinematics_nlp



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from typing import List, NamedTuple


class ComputedTorqueFixedBaseContext(NamedTuple):
    name: str
    kp: List[float]
    ki: List[float]
    kd: List[float]
    urdf: str
    joints: List[str]
    gravity: List[float]

    def to_xml(self) -> str:
        xml = f"""
        <sdf version='1.7'>
            <name>foo</name>
            <test arg='boh'/>
            <controller name="{self.name}">
                <kp>{self._to_str(self.kp)}</kp>
                <ki>{self._to_str(self.ki)}</ki>
                <kd>{self._to_str(self.kd)}</kd>
                <urdf>{self.urdf}</urdf>
                <joints>{self._to_str(self.joints)}</joints>
                <gravity>{self._to_str(self.gravity)}</gravity>
            </controller>
            <!--test/-->
        </sdf>
        """

        return xml

    @staticmethod
    def _to_str(input_list: List):

        list_serialized = ""

        for index, element in enumerate(input_list[:-1]):
            list_serialized += str(element) + " "

        list_serialized += str(input_list[-1])

        return list_serialized


# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
from typing import Tuple
import gym_ignition_models
from gym_ignition.utils import misc
from scenario import gazebo as scenario


def id_gazebo_fn(val: Tuple[float, float, float]):
    if isinstance(val, tuple) and len(val) == 3:
        return f"step={val[0]}_rtf={val[1]}_iters={val[2]}"


@pytest.fixture(scope="function")
def gazebo_fixture(request):
    """
    Return an instance of the GazeboSimulator ensuring that it
    is closed even in case of failure.

    Example:

        from ..common.utils import gazebo_fixture as gazebo

        @pytest.mark.parametrize("gazebo", [(0.001, 1.0, 1)], indirect=True)
        def test_foo(gazebo: scenario.gazebo.GazeboSimulator):

            assert gazebo.initialize()
            # ...

        @pytest.mark.parametrize("gazebo, name",
                                 [((0.001, 1.0, 1), "name_1"),
                                  ((0.001, 1.0, 1), "name_2"),
                                  ((0.001, 1.0, 1), "name_3")],
                                  indirect="gazebo")
        def test_foo(gazebo: scenario.gazebo.GazeboSimulator, name: str):

            model_name = name
            assert gazebo.initialize()
            # ...
    """

    step_size, rtf, iterations = request.param
    gazebo = scenario.GazeboSimulator(step_size, rtf, iterations)

    yield gazebo

    gazebo.close()


@pytest.fixture(scope="function")
def default_world_fixture(request):
    """
    Initialize a default world with ground and physics.

    Example:

        from ..common.utils import default_world_fixture as default_world

        @pytest.mark.parametrize("default_world", [(0.001, 1.0, 1)], indirect=True)
        def test_foo(default_world: Tuple[scenario.GazeboSimulator, scenario.World]):

            # Get the simulator and the world
            gazebo, world = default_world

            # ...
    """

    step_size, rtf, iterations = request.param
    gazebo = scenario.GazeboSimulator(step_size, rtf, iterations)

    assert gazebo.initialize()

    world = gazebo.get_world().to_gazebo()
    assert world.insert_model(gym_ignition_models.get_model_file("ground_plane"))
    assert world.set_physics_engine(scenario.PhysicsEngine_dart)

    yield gazebo, world

    gazebo.close()


def get_multi_world_sdf_file() -> str:

    multi_world_sdf_string = f"""
    <?xml version="1.0" ?>
    <sdf version="1.6">
        <world name="world1">
        </world>
        <world name="world2">
        </world>
    </sdf>"""

    multi_world_sdf = misc.string_to_file(multi_world_sdf_string)
    return multi_world_sdf


def get_cube_urdf_string() -> str:

    mass = 5.0
    edge = 0.2
    i = 1 / 12 * mass * (edge ** 2 + edge ** 2)
    cube_urdf = f"""
    <robot name="cube_robot" xmlns:xacro="http://www.ros.org/wiki/xacro">
        <link name="cube">
            <inertial>
              <origin rpy="0 0 0" xyz="0 0 0"/>
              <mass value="{mass}"/>
              <inertia ixx="{i}" ixy="0" ixz="0" iyy="{i}" iyz="0" izz="{i}"/>
            </inertial>
            <visual>
              <geometry>
                <box size="{edge} {edge} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 0 0"/>
            </visual>
            <collision>
              <geometry>
                <box size="{edge} {edge} {edge}"/>
              </geometry>
              <origin rpy="0 0 0" xyz="0 0 0"/>
            </collision>
        </link>
    </robot>"""

    return cube_urdf


def get_cube_urdf() -> str:

    model_file = misc.string_to_file(get_cube_urdf_string())
    return model_file


def get_empty_world_sdf() -> str:

    world_sdf_string = scenario.get_empty_world()

    world_file = misc.string_to_file(world_sdf_string)
    return world_file



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.gympp

import gym
from scenario import gazebo as scenario

try:
    import gympp_bindings as bindings
except ImportError:
    pytest.skip("gympp bindings not found", allow_module_level=True)


# Set verbosity
scenario.set_verbosity(4)


def test_create_cpp_environment():

    env = gym.make("CartPoleDiscrete-Gympp-v0")
    assert env

    env.seed(42)

    action = env.action_space.sample()
    assert isinstance(action, int)

    observation = env.observation_space.sample()
    assert observation.size > 0

    observation = env.reset()
    assert observation.size > 0

    state, reward, done, _ = env.step(action)
    assert state.size > 0



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy as np
from . import utils
from gym_ignition.robots import sim
import gym_ignition_models as models
from gym_ignition.utils.inverse_kinematics_nlp import TargetType
from gym_ignition.base.robot.robot_joints import JointControlMode, PID
from gym_ignition.utils.inverse_kinematics_nlp import InverseKinematicsNLP

# From: https://github.com/mkrizmancic/franka_gazebo/blob/master/config/default.yaml
# These gains work well with a physics step of 1000 Hz.
pid_gains_1000_hz = {
    'panda_joint1': PID(p=50, i=0, d=20),
    'panda_joint2': PID(p=10000, i=0, d=500),
    'panda_joint3': PID(p=100, i=0, d=10),
    'panda_joint4': PID(p=1000, i=0, d=50),
    'panda_joint5': PID(p=100, i=0, d=10),
    'panda_joint6': PID(p=100, i=0, d=10),
    'panda_joint7': PID(p=10, i=0.5, d=0.1),
    'panda_finger_joint1': PID(p=100, i=0, d=50),
    'panda_finger_joint2': PID(p=100, i=0, d=50),
}


def test_inverse_kinematics():
    # Get the simulator
    physics_rate = 1000.0
    gazebo = utils.Gazebo(physics_rate=physics_rate)

    # Get the robot
    panda = sim.gazebo.panda.PandaRobot(gazebo=gazebo.simulator)

    # Set the PID gains
    for joint_name, pid in pid_gains_1000_hz.items():
        ok_pid = panda.set_joint_pid(joint_name=joint_name, pid=pid)
        assert ok_pid

    # Configure the PID controllers step time
    ok_dt = panda.set_dt(step_size=(1.0 / physics_rate))
    assert ok_dt

    # Get the controlled joints (no fingers)
    controlled_joints = [j for j in panda.joint_names() if "_finger_" not in j]

    # Control the robot in position
    for joint_name in panda.joint_names():
        ok_mode = panda.set_joint_control_mode(joint_name, JointControlMode.POSITION)
        assert ok_mode, f"Failed to control joint {joint_name} in position"

    # Create the IK object
    ik = InverseKinematicsNLP(urdf_filename=models.get_model_file("panda"),
                              considered_joints=controlled_joints)

    # Initialize IK
    ik.initialize(verbosity=1,
                  cost_tolerance=1e-10,
                  floating_base=False)

    # There should be no active targets
    assert len(ik.get_active_target_names()) == 0

    # Add the cartesian target of the end effector
    end_effector = "end_effector_frame"
    ik.add_target(frame_name=end_effector,
                  target_type=TargetType.POSE)

    assert set(ik.get_active_target_names()) == {end_effector}
    assert ik.get_target_data(target_name=end_effector).type == TargetType.POSE

    target_ee_position = np.array([0.3, 0, 0.8])
    target_ee_quaternion = np.array([0.0, 0.0, 0.0, 1.0])

    # Set the target transform
    ik.update_transform_target(target_name=end_effector,
                               position=target_ee_position,
                               quaternion=target_ee_quaternion)

    # Solve the IK problem
    ik.solve()

    # Get the solution
    ik_solution = ik.get_solution()

    assert ik_solution.joint_configuration.size == len(controlled_joints)
    assert np.allclose(ik_solution.base_position, np.array([0.0, 0, 0]))
    assert np.allclose(ik_solution.base_quaternion, np.array([1.0, 0, 0, 0]))

    # Set the desired joint configuration
    for joint_name, position in zip(controlled_joints, ik_solution.joint_configuration):
        ok_pos = panda.set_joint_position(joint_name, position)
        assert ok_pos

    for _ in range(1000):
        gazebo.step()

    # Get the end effector pose
    ee_position, ee_quaternion = panda.link_pose(link_name=end_effector)

    assert np.allclose(target_ee_position, ee_position, atol=0.02)
    assert np.allclose(target_ee_quaternion, ee_quaternion, atol=0.03)

    gazebo.close()



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc


class ModelRandomizer(abc.ABC):

    @abc.abstractmethod
    def randomize_model(self) -> str:
        """
        Randomize the model.

        Return:
            A string with the randomized model.
        """
        pass

    def seed_model_randomizer(self, seed: int) -> None:
        """
        Seed the randomizer to ensure reproducibility.

        Args:
            seed: The seed number.
        """
        pass



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.gympp

import os
import numpy as np
from pathlib import Path
import gym_ignition_models
from gym_ignition.utils import misc
from scenario import gazebo as scenario
from gym_ignition.experimental import gympp

try:
    import gympp_bindings as bindings
except ImportError:
    pytest.skip("gympp bindings not found", allow_module_level=True)


# Set verbosity
scenario.set_verbosity(4)


def test_metadata():

    md = bindings.PluginMetadata()
    assert not md.is_valid()

    environment_name = "EnvironmentName"
    md.set_environment_name(environment_name)
    assert md.get_environment_name() == environment_name

    library_name = "plugin_foo"
    md.set_library_name(library_name)
    assert md.get_library_name() == library_name

    class_name = "foo::bar::GymppPlugin"
    md.set_class_name(class_name)
    assert md.get_class_name() == class_name

    model_name = "RobotModel.sdf"
    md.set_model_file_name(model_name)
    assert md.get_model_file_name() == model_name

    world_name = "Environment.world"
    md.set_world_file_name(world_name)
    assert md.get_world_file_name() == world_name

    agent_rate = 1000
    md.set_agent_rate(agent_rate)
    assert md.get_agent_rate() == agent_rate

    real_time_factor = 1E9
    max_physics_step_size = 0.001
    md.set_physics_data(bindings.PhysicsData(real_time_factor, max_physics_step_size))
    physics_data = md.get_physics_data()
    assert physics_data.rtf == real_time_factor
    assert physics_data.max_step_size == max_physics_step_size


def test_gym_factory():

    # Get the factory
    factory = bindings.GymFactory.Instance()

    # Register a plugin with empty metadata
    md = bindings.PluginMetadata()
    assert not factory.register_plugin(md)

    # Get a not registered environment
    env = factory.make("foo")
    assert not env

    # Check that the CartPole plugin exists
    assert "IGN_GAZEBO_SYSTEM_PLUGIN_PATH" in os.environ, \
        "Variable IGN_GAZEBO_SYSTEM_PLUGIN_PATH not set in the environment"
    directories = os.environ['IGN_GAZEBO_SYSTEM_PLUGIN_PATH'].split(os.pathsep)

    found = False
    for directory in directories:
        matching_plugins = list(Path(directory).glob("*CartPole*"))
        found = found or (len(matching_plugins) is not 0)
    assert found

    world_file = misc.string_to_file(gympp.get_empty_world())
    cartpole_urdf = gym_ignition_models.get_model_file("cartpole")

    # Create the metadata
    md = bindings.PluginMetadata()
    md.set_environment_name("CartPole")
    md.set_library_name("CartPolePlugin")
    md.set_class_name("gympp::plugins::CartPole")
    md.set_world_file_name(world_file)
    md.set_model_file_name(cartpole_urdf)
    md.set_agent_rate(1000)
    md.set_physics_data(bindings.PhysicsData(1.0, 0.001))
    action_space_md = bindings.SpaceMetadata()
    action_space_md.set_type(bindings.SpaceType_discrete)
    action_space_md.set_dimensions([2])
    observation_space_md = bindings.SpaceMetadata()
    observation_space_md.set_type(bindings.SpaceType_box)
    max_float = float(np.finfo(np.float32).max)
    observation_space_md.set_low_limit([-2.5, -max_float, -24, -max_float])
    observation_space_md.set_high_limit([2.5, max_float, 24, max_float])
    md.set_action_space_metadata(action_space_md)
    md.set_observation_space_metadata(observation_space_md)
    assert md.is_valid()

    # Register the metadata
    assert factory.register_plugin(md)

    # Get the environment
    env = factory.make("CartPole")
    assert env

    # Get the gazebo wrapper
    gazebo = bindings.env_to_gazebo_simulator(env)
    assert gazebo

    # Get the ignition environment
    ign_env = bindings.env_to_gazebo_environment(env)
    assert ign_env

    # Use the environment
    env.reset()
    action = env.action_space.sample()
    state_opt = env.step(action)
    assert state_opt.has_value()

    state = state_opt.value()
    assert list(state.observation.get_buffer_d())

    observation = list(state.observation.get_buffer_d())
    assert len(observation) == 4

    # Try to register another environment
    assert factory.register_plugin(md)

    # Try to get another environment
    env2 = factory.make("CartPole")
    assert env2, "Failed to create CartPoleIgnition environment from the factory"
    assert env != env2, "Environment created from the factory are the same"
    env2.reset()
    env2.step(action)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from scenario import gazebo as scenario
from gym_ignition.randomizers.base import physics


class DART(physics.PhysicsRandomizer):
    """
    Class that configures the Ignition Gazebo physics with the DART physics engine and
    no randomization.
    """

    def __init__(self, seed: int = None):

        super().__init__()

        if seed is not None:
            self.seed_physics_randomizer(seed=seed)

    def randomize_physics(self, world: scenario.World) -> None:

        # Insert the physics
        ok_physics = world.set_physics_engine(scenario.PhysicsEngine_dart)

        if not ok_physics:
            raise RuntimeError("Failed to insert the physics plugin")



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from scenario import gazebo as scenario


class PhysicsRandomizer(abc.ABC):
    """
    Abstract class that provides the machinery for randomizing physics in a Ignition
    Gazebo simulation.

    Args:
        randomize_after_rollouts_num: defines after many rollouts physics should be
        randomized (i.e. the amount of times :py:meth:`gym.Env.reset` is called).
    """

    def __init__(self, randomize_after_rollouts_num: int = 0):

        self._rollout_counter = randomize_after_rollouts_num
        self.randomize_after_rollouts_num = randomize_after_rollouts_num

    @abc.abstractmethod
    def randomize_physics(self, world: scenario.World) -> None:
        """
        Method that insert and configures the physics of a world.

        By default this method loads a plugin that uses DART with no randomizations.
        Randomizing physics engine parameters or changing physics engine backend could be
        done by redefining this method and passing it to
        :py:class:`~gym_ignition.runtimes.gazebo_runtime.GazeboRuntime`.

        Args:
            world: A world object without physics.
        """
        pass

    def seed_physics_randomizer(self, seed: int) -> None:
        """
        Seed the randomizer to ensure reproducibility.

        Args:
            seed: The seed number.
        """
        pass

    def increase_rollout_counter(self) -> None:
        """
        Increase the rollouts counter.
        """

        if self.randomize_after_rollouts_num != 0:
            assert self._rollout_counter != 0
            self._rollout_counter -= 1

    def physics_expired(self) -> bool:
        """
        Checks if the physics needs to be randomized.

        Return:
            True if the physics has expired, false otherwise.
        """

        if self.randomize_after_rollouts_num == 0:
            return False

        if self._rollout_counter == 0:
            self._rollout_counter = self.randomize_after_rollouts_num
            return True

        return False



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import os
import platform
import subprocess
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):
    """
    Custom setuptool extension to define parameters for configuring CMake.
    """

    def __init__(self,
                 name: str,
                 cmake_configuration: str,
                 source_dir: str = ''):
        Extension.__init__(self, name, sources=[])
        self.source_dir = os.path.abspath(source_dir)
        self.cmake_configuration = cmake_configuration


class BuildExtension(build_ext):
    """
    Setuptool build extension handler.
    It processes all the extensions listed in the 'ext_modules' entry.
    """

    def run(self):
        try:
            _ = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        if len(self.extensions) != 1 or not \
                isinstance(self.extensions[0], CMakeExtension):
            raise RuntimeError("This class can only build one CMakeExtension object")

        if platform.system() != "Linux":
            raise RuntimeError("Only Linux is currently supported")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        # In editable mode, do not compile C++.
        # Developers should install the project through regular CMake.
        # Note that here we use the 'inplace' variable that it switched on by the
        # pip '--editable' option.
        if self.inplace:
            print("Editable install recognized. The CMake project will not be installed.")
            return

        # Get the temporary external build directory.
        # On Ubuntu it will be '/tmp/pip-req-build-<randomseq>/build/linux-x86_64-3.6'.
        ext_dir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # Shared CMake arguments
        cmake_args = [
            f"-DCMAKE_INSTALL_PREFIX:PATH={ext_dir}",
        ]

        # Apply the selected CMake build type
        cfg = ext.cmake_configuration
        build_args = ['--config', cfg]

        # Handle differences in OS and CMake generators
        if platform.system() == "Windows":
            # Refer to https://github.com/pybind/cmake_example/blob/master/setup.py
            raise NotImplementedError
        elif platform.system() == "Darwin":
            raise NotImplementedError
        elif platform.system() == "Linux":
            cmake_args += [
                f"-GNinja",
                f"-DCMAKE_BUILD_TYPE={cfg}",
            ]
            install_target = "install"
        else:
            raise RuntimeError(f"Unsupported '{platform.system()}' platform")

        # Commands to be executed
        configure_command = ['cmake', ext.source_dir] + cmake_args
        build_command = ['cmake', '--build', '.'] + build_args
        install_command = ['cmake', '--build', '.', '--target', install_target]

        print("")
        print(f"Build folder: {self.build_temp}")
        print("")
        print(f"1) Configuring: {configure_command}")
        print(f"2) Building   : {build_command}")
        print(f"3) Installing : {install_command}")
        print("")

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        # Call CMake
        subprocess.check_call(configure_command, cwd=self.build_temp)
        subprocess.check_call(build_command, cwd=self.build_temp)
        subprocess.check_call(install_command, cwd=self.build_temp)


# Read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gym-ignition',
    author="Diego Ferigo",
    author_email="diego.ferigo@iit.it",
    description="Gym-Ignition: A toolkit for developing OpenAI Gym environments "
                "simulated with Ignition Gazebo.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/robotology/gym-ignition',
    keywords="openai gym reinforcement learning environment gazebo robotics ignition",
    license="LGPL",
    platforms='any',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Topic :: Games/Entertainment :: Simulation",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Framework :: Robot Framework",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
    ],
    use_scm_version={
        'local_scheme': 'dirty-tag',
    },
    setup_requires=['setuptools_scm'],
    python_requires='>=3.6',
    install_requires=[
        'gym >= 0.13.1',
        'numpy',
        'gym_ignition_models',
        'lxml',
    ],
    extras_require=dict(
        website=[
            "sphinx",
            "sphinx-rtd-theme",
            "sphinx-autodoc-typehints",
            "sphinx_fontawesome",
            "sphinx-multiversion",
            "breathe",
        ],
    ),
    packages=find_packages("python"),
    package_dir={'': "python"},
    ext_modules=[CMakeExtension(name='InstallAllTargets', cmake_configuration='PyPI')],
    cmdclass={
        'build_ext': BuildExtension,
    },
    zip_safe=False,
)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy as np
import gym_ignition_models
from gym_ignition.utils import misc
from gym_ignition.experimental import gympp
from gym_ignition.experimental.gympp import gympp_env


class CartPoleDiscrete(gympp_env.GymppEnv):

    def __init__(self):

        # Initialize the parent class
        super().__init__()

    @property
    def _plugin_metadata(self) -> "gympp_bindings.PluginMetadata":

        # Lazy module import
        import gympp_bindings as bindings

        # Get the empty world file
        empty_world_sdf = misc.string_to_file(gympp.get_empty_world())

        # Configure ignition environment
        md = bindings.PluginMetadata()
        md.set_environment_name("CartPole")
        md.set_library_name("CartPolePlugin")
        md.set_class_name("gympp::plugins::CartPole")
        md.set_world_file_name(empty_world_sdf)
        md.set_model_file_name(gym_ignition_models.get_model_file("cartpole"))
        md.set_agent_rate(1000)

        real_time_factor = 1E9
        max_physics_step_size = 0.001
        md.set_physics_data(bindings.PhysicsData(real_time_factor, max_physics_step_size))

        # Configure the action space
        action_space_md = bindings.SpaceMetadata()
        action_space_md.set_type(bindings.SpaceType_discrete)
        action_space_md.set_dimensions([2])

        # Configure the observation space
        observation_space_md = bindings.SpaceMetadata()
        observation_space_md.set_type(bindings.SpaceType_box)
        max_float = float(np.finfo(np.float32).max)
        observation_space_md.set_low_limit([-2.5, -max_float, -24, -max_float])
        observation_space_md.set_high_limit([2.5, max_float, 24, max_float])

        # Store the spaces information
        md.set_action_space_metadata(action_space_md)
        md.set_observation_space_metadata(observation_space_md)

        # Return the metadata
        assert md.is_valid(), "The plugin metadata object is not valid"
        return md



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from gym_ignition import base


class TaskRandomizer(abc.ABC):

    @abc.abstractmethod
    def randomize_task(self,
                       task: base.task.Task,
                       **kwargs) -> None:
        """
        Randomize a :py:class:`~gym_ignition.base.task.Task` instance.

        Args:
            task: the task to randomize.

        Note:
            Note that each task has a :py:attr:`~gym_ignition.base.task.Task.world`
            property that provides access to the simulated
            :py:class:`scenario.core.World`.
        """
        pass

    def seed_task_randomizer(self, seed: int) -> None:
        """
        Seed the randomizer to ensure reproducibility.

        Args:
            seed: The seed number.
        """
        pass



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import os
import numpy as np
from enum import Enum, auto
from typing import List, Dict, Optional, Union, NamedTuple


class TargetType(Enum):
    POSITION = auto()
    ROTATION = auto()
    POSE = auto()


class TransformTargetData(NamedTuple):
    position: np.ndarray
    quaternion: np.ndarray


class TargetData(NamedTuple):
    type: TargetType
    weight: float
    data: Union[np.ndarray, TransformTargetData]


class IKSolution(NamedTuple):
    joint_configuration: np.ndarray
    base_position: np.ndarray = np.array([0.0, 0.0, 0.0])
    base_quaternion: np.ndarray = np.array([1.0, 0.0, 0.0, 0.0])


class RotationParametrization(Enum):
    QUATERNION = auto()
    ROLL_PITCH_YAW = auto()

    def to_idyntree(self):
        import iDynTree

        if self.value == RotationParametrization.QUATERNION.value:
            return iDynTree.InverseKinematicsRotationParametrizationQuaternion
        elif self.value == RotationParametrization.ROLL_PITCH_YAW.value:
            return iDynTree.InverseKinematicsRotationParametrizationRollPitchYaw
        else:
            raise ValueError(self.value)


class TargetResolutionMode(Enum):
    TARGET_AS_CONSTRAINT_FULL = auto()
    TARGET_AS_CONSTRAINT_NONE = auto()
    TARGET_AS_CONSTRAINT_POSITION = auto()
    TARGET_AS_CONSTRAINT_ROTATION = auto()

    def to_idyntree(self):
        import iDynTree

        if self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_FULL.value:
            return iDynTree.InverseKinematicsTreatTargetAsConstraintFull
        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_NONE.value:
            return iDynTree.InverseKinematicsTreatTargetAsConstraintNone
        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_POSITION.value:
            return iDynTree.InverseKinematicsTreatTargetAsConstraintPositionOnly
        elif self.value == TargetResolutionMode.TARGET_AS_CONSTRAINT_ROTATION.value:
            return iDynTree.InverseKinematicsTreatTargetAsConstraintRotationOnly
        else:
            raise ValueError(self.value)


class InverseKinematicsNLP:

    def __init__(self, urdf_filename: str, considered_joints: List[str] = None) -> None:

        import iDynTree

        self._base_frame: Optional[str] = None
        self._urdf_filename: str = urdf_filename
        self._targets_data: Dict[str, TargetData] = dict()
        self._ik: Optional[iDynTree.InverseKinematics] = None
        self._considered_joints: List[str] = considered_joints

    # ======================
    # INITIALIZATION METHODS
    # ======================

    def initialize(self,
                   rotation_parametrization: RotationParametrization =
                   RotationParametrization.ROLL_PITCH_YAW,
                   target_mode: TargetResolutionMode =
                   TargetResolutionMode.TARGET_AS_CONSTRAINT_NONE,
                   cost_tolerance: float = 1E-8,
                   max_iterations: int = 1000,
                   base_frame: str = None,
                   floating_base: bool = False,
                   verbosity: int = 1) -> None:

        # Lazy import iDynTree
        import iDynTree

        # Create the IK object
        self._ik = iDynTree.InverseKinematics()

        # Load the iDynTree model and get the loader
        model_loader: iDynTree.ModelLoader = self._get_model_loader(
            urdf=self._urdf_filename, considered_joints=self._considered_joints)

        # Get the model
        model = model_loader.model()

        # If all joints are enabled, get the list of joint names in order to know the
        # serialization of the IK solution
        if self._considered_joints is None:
            self._considered_joints = []

            for joint_idx in range(model.getNrOfJoints()):
                self._considered_joints.append(model.getJointName(joint_idx))

        # Set the model
        ok_model = self._ik.setModel(model)

        if not ok_model:
            raise RuntimeError("Failed to set the model in the IK object")

        # Configure IK
        self._ik.setVerbosity(verbosity)
        self._ik.setMaxIterations(max_iterations)
        self._ik.setCostTolerance(cost_tolerance)
        self._ik.setDefaultTargetResolutionMode(target_mode.to_idyntree())
        self._ik.setRotationParametrization(rotation_parametrization.to_idyntree())

        # Optionally change the base frame
        if base_frame is not None:
            # Store the frame of the base link
            self._base_frame = base_frame

            # Change the base frame
            ok_base = self._ik.setFloatingBaseOnFrameNamed(base_frame)

            if not ok_base:
                raise RuntimeError("Failed to change floating base frame")
        else:
            self._base_frame = model.getLinkName(model.getDefaultBaseLink())

        if not floating_base:
            # Add a frame constraint for the base
            self.add_frame_transform_constraint(frame_name=self._base_frame,
                                                position=np.array([0.0, 0, 0]),
                                                quaternion=np.array([1.0, 0, 0, 0]))

    # ==============
    # TARGET METHODS
    # ==============

    def add_target(self,
                   frame_name: str,
                   target_type: TargetType,
                   weight: float = 1.0,
                   as_constraint: bool = False) -> None:
        # Lazy import iDynTree
        import iDynTree as iDyn

        if target_type == TargetType.ROTATION:
            # Add the target
            ok_target = self._ik.addRotationTarget(frame_name,
                                                   iDyn.Rotation.Identity(),
                                                   weight)

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(type=TargetType.ROTATION,
                                                        weight=weight,
                                                        data=np.array([1.0, 0, 0, 0]))

        elif target_type == TargetType.POSITION:
            # Add the target
            ok_target = self._ik.addPositionTarget(frame_name,
                                                   iDyn.Position_Zero(),
                                                   weight)

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(type=TargetType.POSITION,
                                                        weight=weight,
                                                        data=np.array([0.0, 0, 0]))

        elif target_type == TargetType.POSE:
            # Add the target
            ok_target = self._ik.addTarget(frame_name,
                                           iDyn.Transform.Identity(),
                                           weight)

            # Create the transform target data
            target_data = TransformTargetData(position=np.array([0.0, 0, 0]),
                                              quaternion=np.array([1., 0, 0, 0]))

            # Initialize the target data buffers
            self._targets_data[frame_name] = TargetData(type=TargetType.POSE,
                                                        weight=weight,
                                                        data=target_data)

        else:
            raise ValueError(target_type)

        if not ok_target:
            raise RuntimeError(f"Failed to add target for frame '{frame_name}'")

        if as_constraint:
            if target_type == TargetType.ROTATION:
                constraint = iDyn.InverseKinematicsTreatTargetAsConstraintRotationOnly
            elif target_type == TargetType.POSITION:
                constraint = iDyn.InverseKinematicsTreatTargetAsConstraintPositionOnly
            else:
                assert target_type == TargetType.POSE
                constraint = iDyn.InverseKinematicsTreatTargetAsConstraintFull

            ok_constraint = self._ik.setTargetResolutionMode(frame_name, constraint)

            if not ok_constraint:
                raise RuntimeError(f"Failed to set target '{frame_name}' as constraint")

    def update_position_target(self, target_name: str, position: np.ndarray) -> None:
        if target_name not in self.get_active_target_names():
            raise ValueError(f"Target '{target_name}' was never added")

        # Create the iDynTree position
        p = self._to_idyntree_position(position=position)

        # Update the target inside IK
        ok_target = self._ik.updateTarget(target_name, p)

        if not ok_target:
            raise RuntimeError(f"Failed to update position of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Update the target data
        self._targets_data[target_name] = TargetData(type=TargetType.ROTATION,
                                                     weight=weight,
                                                     data=position)

    def update_rotation_target(self, target_name: str, quaternion: np.ndarray) -> None:
        if target_name not in self.get_active_target_names():
            raise ValueError(f"Target '{target_name}' was never added")

        # Create the iDynTree rotation matrix
        R = self._to_idyntree_rotation(quaternion=quaternion)

        ok_target = self._ik.updateRotationTarget(target_name, R)

        if not ok_target:
            raise RuntimeError(f"Failed to update rotation of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Update the target data
        self._targets_data[target_name] = TargetData(type=TargetType.ROTATION,
                                                     weight=weight,
                                                     data=quaternion)

    def update_transform_target(self,
                                target_name: str,
                                position: np.ndarray,
                                quaternion: np.ndarray) -> None:
        if target_name not in self.get_active_target_names():
            raise ValueError(f"Target '{target_name}' was never added")

        # Create the iDynTree transform
        H = self._to_idyntree_transform(position=position, quaternion=quaternion)

        # Update the target inside IK
        ok_target = self._ik.updateTarget(target_name, H)

        if not ok_target:
            raise RuntimeError(f"Failed to update transform of target '{target_name}'")

        # Get the configured weight
        weight = self._targets_data[target_name].weight

        # Create the transform target data
        transform_data = TransformTargetData(position=position, quaternion=quaternion)

        # Update the target data
        self._targets_data[target_name] = TargetData(type=TargetType.POSE,
                                                     weight=weight,
                                                     data=transform_data)

    # =============
    # FRAME METHODS
    # =============

    def add_frame_transform_constraint(self,
                                       frame_name: str,
                                       position: np.ndarray,
                                       quaternion: np.ndarray) -> None:
        # Create the transform
        transform = self._to_idyntree_transform(position=position, quaternion=quaternion)

        # Add the target
        ok_constraint = self._ik.addFrameConstraint(frame_name, transform)

        if not ok_constraint:
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def add_frame_position_constraint(self,
                                      frame_name: str,
                                      position: np.ndarray) -> None:
        # Create the position
        position = self._to_idyntree_position(position=position)

        # Add the target
        ok_constraint = self._ik.addFramePositionConstraint(frame_name, position)

        if not ok_constraint:
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def add_frame_rotation_constraint(self,
                                      frame_name: str,
                                      quaternion: np.ndarray) -> None:
        # Create the position
        rotation = self._to_idyntree_rotation(quaternion=quaternion)

        # Add the target
        ok_constraint = self._ik.addFrameRotationConstraint(frame_name, rotation)

        if not ok_constraint:
            raise RuntimeError(f"Failed to add constraint on frame '{frame_name}'")

    def deactivate_frame_constraint(self, frame_name: str) -> None:
        if not self._ik.isFrameConstraintActive(frame_name):
            raise RuntimeError(f"Constraint on frame '{frame_name}' not active")

        ok_removed = self._ik.deactivateFrameConstraint(frame_name)

        if not ok_removed:
            raise RuntimeError(f"Failed to deactivate constraint on frame '{frame_name}")

    # ===================
    # IK SOLUTION METHODS
    # ===================

    def solve(self) -> None:

        ok_solved = self._ik.solve()

        if not ok_solved:
            raise RuntimeError("Failed to solve IK")

        # Initialize next solver call
        self._warm_start_with_last_solution()

    def warm_start_from(self, warm_start_solution: IKSolution) -> None:

        if warm_start_solution.joint_configuration.size != len(self._considered_joints):
            raise RuntimeError(
                "The joint configuration does not match the number of considered joints")

        H = self._to_idyntree_transform(position=warm_start_solution.base_position,
                                        quaternion=warm_start_solution.base_quaternion)
        q = self._to_idyntree_dyn_vector(array=warm_start_solution.joint_configuration)

        # Warm start the solver
        ok_init = self._ik.setFullJointsInitialCondition(H, q)

        if not ok_init:
            raise RuntimeError("Failed to warm start the IK solver")

    # =======
    # GETTERS
    # =======

    def get_base_frame(self) -> str:
        return self._base_frame

    def get_available_target_names(self) -> List[str]:

        # Get the model
        model = self._ik.model()

        # Note that also frames (modeled as fake links) are available targets
        link_names = []

        for link_index in range(model.getNrOfLinks()):
            link_names.append(model.getLinkName(link_index))

        return link_names

    def get_active_target_names(self) -> List[str]:
        return list(self._targets_data.keys())

    def get_target_data(self, target_name: str) -> TargetData:
        return self._targets_data[target_name]

    def get_solution(self) -> IKSolution:

        # Lazy import iDynTree
        import iDynTree

        # Initialize buffers
        base_transform = iDynTree.Transform.Identity()
        joint_positions = iDynTree.VectorDynSize(self._ik.model().getNrOfJoints())

        # Get the solution
        self._ik.getFullJointsSolution(base_transform, joint_positions)

        # Convert to numpy objects
        joint_positions = joint_positions.toNumPy()
        base_position = base_transform.getPosition().toNumPy()
        base_quaternion = base_transform.getRotation().asQuaternion().toNumPy()

        return IKSolution(base_position=base_position,
                          base_quaternion=base_quaternion,
                          joint_configuration=joint_positions)

    # ===============
    # PRIVATE METHODS
    # ===============

    @staticmethod
    def _to_idyntree_dyn_vector(array: np.ndarray):
        import iDynTree
        dyn_vector = iDynTree.VectorDynSize()
        dyn_vector = dyn_vector.FromPython(array)

        return dyn_vector

    @staticmethod
    def _to_idyntree_position(position: np.ndarray):
        if position.size != 3:
            raise ValueError("The position must have 3 elements")

        import iDynTree
        p = iDynTree.Position()
        for i in range(3):
            p.setVal(i, position[i])

        return p

    @staticmethod
    def _to_idyntree_rotation(quaternion: np.ndarray):
        if quaternion.size != 4:
            raise ValueError("The quaternion must have 4 elements")

        import iDynTree
        quat = iDynTree.Vector4()
        quat = quat.FromPython(quaternion)

        R = iDynTree.Rotation()
        R.fromQuaternion(quat)

        return R

    @staticmethod
    def _to_idyntree_transform(position: np.ndarray, quaternion: np.ndarray):
        p = InverseKinematicsNLP._to_idyntree_position(position=position)
        R = InverseKinematicsNLP._to_idyntree_rotation(quaternion=quaternion)

        import iDynTree
        H = iDynTree.Transform()
        H.setPosition(p)
        H.setRotation(R)

        return H

    @staticmethod
    def _get_model_loader(urdf: str, considered_joints: List[str] = None):
        # Lazy import iDynTree
        import iDynTree

        # Get the model loader
        model_loader = iDynTree.ModelLoader()

        if not os.path.exists(urdf):
            raise FileNotFoundError(urdf)

        # Load the model
        if considered_joints:
            ok_load = model_loader.loadReducedModelFromFile(urdf, considered_joints)
        else:
            ok_load = model_loader.loadModelFromFile(urdf)

        if not ok_load:
            raise RuntimeError("Failed to load model")

        # Due to some SWIG internal, returning the model contained by the loader
        # does not work as expected
        return model_loader

    def _warm_start_with_last_solution(self) -> None:
        last_solution = self.get_solution()
        self.warm_start_from(last_solution)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.gympp

import numpy as np
from numpy import pi
from scenario import gazebo as scenario


try:
    import gympp_bindings as bindings
except ImportError:
    pytest.skip("gympp bindings not found", allow_module_level=True)


# Set verbosity
scenario.set_verbosity(4)


@pytest.fixture
def create_std_vector():

    python_list = [1.0, 2.0, 3.0]
    vector = bindings.VectorD(python_list)

    return vector


@pytest.fixture
def create_space_box_md():

    md = bindings.SpaceMetadata()
    md.set_type(bindings.SpaceType_box)
    max_float = float(np.finfo(np.float32).max)
    md.set_low_limit([-2.5, -max_float, -24, -max_float])
    md.set_high_limit([2.5, max_float, 24, max_float])

    return md


@pytest.fixture
def create_space_discrete_md():

    md = bindings.SpaceMetadata()
    md.set_type(bindings.SpaceType_discrete)
    md.set_dimensions([2])

    return md


def test_vectors():

    python_list = [1.0, 2.0, 3.0]
    vector = bindings.VectorD(python_list)

    for i in range(0, vector.size()-1):
        assert python_list[i] == vector[i]


def test_sample(create_std_vector):

    vector = create_std_vector
    sample = bindings.Sample(vector)

    for i in range(0, vector.size()-1):
        assert sample.get_buffer_d()[i] == vector[i]

    sample.get_buffer_d()[2] = 42
    assert sample.get_d(2).value() == 42
    assert sample.get_buffer_d()[2] == 42


def test_range():

    gympp_range = bindings.Range(-5.0, 10.0)

    assert gympp_range.contains(0)
    assert gympp_range.contains(-5)
    assert gympp_range.contains(10)
    assert not gympp_range.contains(-6)
    assert not gympp_range.contains(10.1)


def test_discrete_space():

    discrete_space = bindings.Discrete(12)

    assert not discrete_space.contains(bindings.Sample([-1]))
    assert not discrete_space.contains(bindings.Sample([13]))

    for n in range(50):
        sample = discrete_space.sample()
        assert sample.get_buffer_i().size() == 1
        assert isinstance(sample.get_buffer_i()[0], int)
        assert discrete_space.contains(sample)


def test_box_space():

    size = 4
    box = bindings.Box(-1, 42, [size])

    # By default the data precision of python list is float. Force double.
    assert box.contains(bindings.Sample(bindings.VectorD([0, pi, 12, 42])))
    assert not box.contains(bindings.Sample(bindings.VectorD([0, pi, 12, 43])))
    assert not box.contains(bindings.Sample(bindings.VectorD([0])))
    assert not box.contains(bindings.Sample(bindings.VectorD([0, pi, 12, 43, 0])))

    for n in range(50):

        sample = box.sample()

        assert sample.get_buffer_d().size() == size
        assert isinstance(sample.get_buffer_d()[0], float)
        assert box.contains(sample)


def test_space_box_metadata(create_space_box_md):

    md = create_space_box_md

    assert md.is_valid()
    assert md.get_type() == bindings.SpaceType_box


def test_space_discrete_metadata(create_space_discrete_md):

    md = create_space_discrete_md

    assert md.is_valid()
    assert md.get_type() == bindings.SpaceType_discrete
